#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['velin']

package_data = \
{'': ['*']}

install_requires = \
['numpydoc', 'pygments', 'black', 'there']

entry_points = \
{'console_scripts': ['velin = velin:main']}

setup(name='velin',
      version='0.0.6',
      description='French for Vellum',
      author='Matthias Bussonnier',
      author_email='bussonniermatthias@gmail.com',
      url='https://github.com/Carreau/minirst',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.8',
     )
